#include <windows.h>
//#include <maths.h>

WCHAR szClassName[] = L"RadioTestApp";

#define RAD_SZ 16
#define BGRAD_SZ 40

HPEN GetSysColorPen(INT nIndex)
{
     return CreatePen(PS_SOLID, 0, GetSysColor(nIndex));
}

static int UITOOLS_MakeSquareRect(LPRECT src, LPRECT dst)
{
    int Width  = src->right - src->left;
    int Height = src->bottom - src->top;
    int SmallDiam = Width > Height ? Height : Width;

    *dst = *src;

    /* Make it a square box */
    if(Width < Height)      /* SmallDiam == Width */
    {
        dst->top += (Height-Width)/2;
        dst->bottom = dst->top + SmallDiam;
    }
    else if(Width > Height) /* SmallDiam == Height */
    {
        dst->left += (Width-Height)/2;
        dst->right = dst->left + SmallDiam;
    }

   return SmallDiam;
}


static BOOL UITOOLS95_DFC_ButtonRadio(HDC dc, LPRECT r, UINT uFlags)
{
    RECT myr;
    int i;
    int SmallDiam = UITOOLS_MakeSquareRect(r, &myr);
    int BorderShrink = SmallDiam / 16;
    HPEN hpsave;
    HBRUSH hbsave;
    int xc, yc;

    if(BorderShrink < 1) BorderShrink = 1;

    if((uFlags & 0xff) == DFCS_BUTTONRADIOIMAGE)
    {
        FillRect(dc, r, (HBRUSH)GetStockObject(BLACK_BRUSH));
    }

    xc = myr.left + SmallDiam - (SmallDiam / 2);
    yc = myr.top  + SmallDiam - (SmallDiam / 2);

    /* Define bounding box */
    i = 14 * SmallDiam / 16;
    myr.left   = (xc - i) + (i / 2);
    myr.right  =  xc + (i / 2);
    myr.top    = (yc - i) + (i / 2);
    myr.bottom =  yc + (i / 2);

    if((uFlags & 0xff) == DFCS_BUTTONRADIOMASK)
    {
        hbsave = (HBRUSH)SelectObject(dc, GetStockObject(BLACK_BRUSH));
        Ellipse(dc, myr.left, myr.top, myr.right, myr.bottom);
        SelectObject(dc, hbsave);
    }
    else
    {
        if(uFlags & (DFCS_FLAT|DFCS_MONO))
        {
            hpsave = (HPEN)SelectObject(dc, GetSysColorPen(COLOR_WINDOWFRAME));
            hbsave = (HBRUSH)SelectObject(dc, GetSysColorBrush(COLOR_WINDOWFRAME));
            Ellipse(dc, myr.left, myr.top, myr.right, myr.bottom);
            SelectObject(dc, hbsave);
            SelectObject(dc, hpsave);
        }
        else
        {
            /* outer bottom right white section */
            hpsave = (HPEN)SelectObject(dc, GetSysColorPen(COLOR_BTNHIGHLIGHT));
            hbsave = (HBRUSH)SelectObject(dc, GetSysColorBrush(COLOR_BTNHIGHLIGHT));
            Pie(dc, myr.left, myr.top, myr.right+1, myr.bottom+1, myr.left-1, myr.bottom, myr.right+1, myr.top);

            /* outer top left, light grey section */
            SelectObject(dc, GetSysColorPen(COLOR_BTNSHADOW));
            SelectObject(dc, GetSysColorBrush(COLOR_BTNSHADOW));
            Pie(dc, myr.left, myr.top, myr.right+1, myr.bottom+1, myr.right+1, myr.top, myr.left-1, myr.bottom);

            myr.left   += BorderShrink;
            myr.right  -= BorderShrink;
            myr.top    += BorderShrink;
            myr.bottom -= BorderShrink;

            /* inner bottom right, button face section */
            SelectObject(dc, GetSysColorPen(COLOR_3DLIGHT));
            SelectObject(dc, GetSysColorBrush(COLOR_3DLIGHT));
            Pie(dc, myr.left, myr.top, myr.right+1, myr.bottom+1, myr.left-1, myr.bottom, myr.right+1, myr.top);

            /* inner top left, dark grey section */
            SelectObject(dc, GetSysColorPen(COLOR_3DDKSHADOW));
            SelectObject(dc, GetSysColorBrush(COLOR_3DDKSHADOW));
            Pie(dc, myr.left, myr.top, myr.right+1, myr.bottom+1, myr.right+1, myr.top, myr.left-1, myr.bottom);
            SelectObject(dc, hbsave);
            SelectObject(dc, hpsave);
        }

        i = 10 * SmallDiam / 16;
        if (i % 2 != 0) i++;
        myr.left   = xc - i+i/2;
        myr.right  = xc + i/2;
        myr.top    = yc - i+i/2;
        myr.bottom = yc + i/2;
        i= !(uFlags & (DFCS_INACTIVE|DFCS_PUSHED)) ? COLOR_WINDOW : COLOR_BTNFACE;
        hpsave = (HPEN)SelectObject(dc, GetSysColorPen(i));
        hbsave = (HBRUSH)SelectObject(dc, GetSysColorBrush(i));
        Ellipse(dc, myr.left, myr.top, myr.right, myr.bottom);
        SelectObject(dc, hbsave);
        SelectObject(dc, hpsave);
    }

    if(uFlags & DFCS_CHECKED)
    {
        //float t;
        i = 6 * SmallDiam / 16;
        if (i % 2 != 0) i--;
        //t = fmod(i, 2);
        i = i < 1 ? 1 : i;
        myr.left   = (xc - i) + (i / 2);
        myr.right  = xc + i/2;
        myr.top    = (yc - i) + (i / 2);
        myr.bottom = yc + i/2;

        i = uFlags & DFCS_INACTIVE ? COLOR_BTNSHADOW : COLOR_WINDOWTEXT;
        hbsave = (HBRUSH)SelectObject(dc, GetSysColorBrush(i));
        hpsave = (HPEN)SelectObject(dc, GetSysColorPen(i));
        Ellipse(dc, myr.left, myr.top, myr.right, myr.bottom);
        SelectObject(dc, hpsave);
        SelectObject(dc, hbsave);
    }

    /* FIXME: M$ has a Polygon in the center at relative points: */
    /* 0.476, 0.476 (times SmallDiam, SmallDiam) */
    /* 0.476, 0.525 */
    /* 0.500, 0.500 */
    /* 0.500, 0.499 */
    /* when the button is unchecked. The reason for it is unknown. The */
    /* color is COLOR_BTNHIGHLIGHT, although the Polygon gets painted at */
    /* least 3 times (it looks like a clip-region when you see it happen). */
    /* I do not really see a reason why this should be implemented. If you */
    /* have a good reason, let me know. Maybe this is a quirk in the Marlett */
    /* font. */

    return TRUE;
}


static BOOL
UITOOLS95_DrawFrameButton(HDC hdc, LPRECT rc, UINT uState)
{
    switch(uState & 0xff)
    {
    case DFCS_BUTTONRADIOIMAGE:
    case DFCS_BUTTONRADIOMASK:
    case DFCS_BUTTONRADIO:
        return UITOOLS95_DFC_ButtonRadio(hdc, rc, uState);
    }

    return FALSE;
}


BOOL
MyDrawFrameControl(HDC hDC, LPRECT rc, UINT uType, UINT uState)
{
   if (GetMapMode(hDC) != MM_TEXT)
      return FALSE;

   switch(uType)
   {
      case DFC_BUTTON:
         return UITOOLS95_DrawFrameButton(hDC, rc, uState);
   }
   return FALSE;
}




LRESULT CALLBACK
WindowProcedure(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    static RECT rc[12];
    static RECT bgrc[12];

    switch (message)
    {
        case WM_CREATE:
            /* Windows buttons */
            SetRect(&rc[0],          0, 0,     RAD_SZ, RAD_SZ);
            SetRect(&rc[1],     RAD_SZ, 0, RAD_SZ * 2, RAD_SZ);
            SetRect(&rc[2], RAD_SZ * 2, 0, RAD_SZ * 3, RAD_SZ);
            SetRect(&rc[3], RAD_SZ * 3, 0, RAD_SZ * 4, RAD_SZ);
            SetRect(&rc[4], RAD_SZ * 4, 0, RAD_SZ * 5, RAD_SZ);
            SetRect(&rc[5], RAD_SZ * 5, 0, RAD_SZ * 6, RAD_SZ);

            /* ROS buttons */
            SetRect(&rc[6],   RAD_SZ * 7, 0, RAD_SZ *  8, RAD_SZ);
            SetRect(&rc[7],   RAD_SZ * 8, 0, RAD_SZ *  9, RAD_SZ);
            SetRect(&rc[8],   RAD_SZ * 9, 0, RAD_SZ * 10, RAD_SZ);
            SetRect(&rc[9],  RAD_SZ * 10, 0, RAD_SZ * 11, RAD_SZ);
            SetRect(&rc[10], RAD_SZ * 11, 0, RAD_SZ * 12, RAD_SZ);
            SetRect(&rc[11], RAD_SZ * 12, 0, RAD_SZ * 13, RAD_SZ);

            /* large Windows buttons */
            SetRect(&bgrc[0],            0, RAD_SZ + 6,     BGRAD_SZ, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[1],     BGRAD_SZ, RAD_SZ + 6, BGRAD_SZ * 2, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[2], BGRAD_SZ * 2, RAD_SZ + 6, BGRAD_SZ * 3, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[3], BGRAD_SZ * 3, RAD_SZ + 6, BGRAD_SZ * 4, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[4], BGRAD_SZ * 4, RAD_SZ + 6, BGRAD_SZ * 5, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[5], BGRAD_SZ * 5, RAD_SZ + 6, BGRAD_SZ * 6, BGRAD_SZ + RAD_SZ + 6);

            /* large ROS buttons */
            SetRect(&bgrc[6],   BGRAD_SZ * 7, RAD_SZ + 6,  BGRAD_SZ * 8, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[7],   BGRAD_SZ * 8, RAD_SZ + 6,  BGRAD_SZ * 9, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[8],   BGRAD_SZ * 9, RAD_SZ + 6, BGRAD_SZ * 10, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[9],  BGRAD_SZ * 10, RAD_SZ + 6, BGRAD_SZ * 11, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[10], BGRAD_SZ * 11, RAD_SZ + 6, BGRAD_SZ * 12, BGRAD_SZ + RAD_SZ + 6);
            SetRect(&bgrc[11], BGRAD_SZ * 12, RAD_SZ + 6, BGRAD_SZ * 13, BGRAD_SZ + RAD_SZ + 6);

            break;

        case WM_PAINT:
        {
            PAINTSTRUCT ps;
            HDC hdc = BeginPaint(hwnd, &ps);

            /* Windows buttons */
            DrawFrameControl(hdc, &rc[0], DFC_BUTTON, DFCS_BUTTONRADIO);
            DrawFrameControl(hdc, &rc[1], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_CHECKED);
            DrawFrameControl(hdc, &rc[2], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_FLAT);
            DrawFrameControl(hdc, &rc[3], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_INACTIVE);
            DrawFrameControl(hdc, &rc[4], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_MONO);
            DrawFrameControl(hdc, &rc[5], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_PUSHED);

            /* ROS buttons */
            MyDrawFrameControl(hdc,  &rc[6], DFC_BUTTON, DFCS_BUTTONRADIO);
            MyDrawFrameControl(hdc,  &rc[7], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_CHECKED);
            MyDrawFrameControl(hdc,  &rc[8], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_FLAT);
            MyDrawFrameControl(hdc,  &rc[9], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_INACTIVE);
            MyDrawFrameControl(hdc, &rc[10], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_MONO);
            MyDrawFrameControl(hdc, &rc[11], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_PUSHED);

            /* large Windows buttons */
            DrawFrameControl(hdc, &bgrc[0], DFC_BUTTON, DFCS_BUTTONRADIO);
            DrawFrameControl(hdc, &bgrc[1], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_CHECKED);
            DrawFrameControl(hdc, &bgrc[2], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_FLAT);
            DrawFrameControl(hdc, &bgrc[3], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_INACTIVE);
            DrawFrameControl(hdc, &bgrc[4], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_MONO);
            DrawFrameControl(hdc, &bgrc[5], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_PUSHED);

            /* large ROS buttons */
            MyDrawFrameControl(hdc, &bgrc[6], DFC_BUTTON, DFCS_BUTTONRADIO);
            MyDrawFrameControl(hdc, &bgrc[7], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_CHECKED);
            MyDrawFrameControl(hdc, &bgrc[8], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_FLAT);
            MyDrawFrameControl(hdc, &bgrc[9], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_INACTIVE);
            MyDrawFrameControl(hdc, &bgrc[10], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_MONO);
            MyDrawFrameControl(hdc, &bgrc[11], DFC_BUTTON, DFCS_BUTTONRADIO | DFCS_PUSHED);

            EndPaint(hwnd, &ps);
            break;
        }

        case WM_DESTROY:
            PostQuitMessage(0);
            break;
        default:
            return DefWindowProc(hwnd, message, wParam, lParam);
    }

    return 0;
}

int WINAPI WinMain(HINSTANCE hThisInstance,
                   HINSTANCE hPrevInstance,
                   LPSTR lpszArgument,
                   int nCmdShow)
{
    HWND hwnd;
    MSG messages;
    WNDCLASSEXW wincl;

    wincl.hInstance = hThisInstance;
    wincl.lpszClassName = szClassName;
    wincl.lpfnWndProc = WindowProcedure;
    wincl.style = CS_DBLCLKS;
    wincl.cbSize = sizeof(WNDCLASSEX);

    wincl.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wincl.hIconSm = LoadIcon(NULL, IDI_APPLICATION);
    wincl.hCursor = LoadCursor(NULL, IDC_ARROW);
    wincl.lpszMenuName = NULL;
    wincl.cbClsExtra = 0;
    wincl.cbWndExtra = 0;
    wincl.hbrBackground = (HBRUSH)COLOR_BACKGROUND;

    if (!RegisterClassExW(&wincl))
        return 0;

    hwnd = CreateWindowExW(0,
                           szClassName,
                           L"Radio Button Test App",
                           WS_OVERLAPPEDWINDOW,
                           CW_USEDEFAULT,
                           CW_USEDEFAULT,
                           550,
                           100,
                           HWND_DESKTOP,
                           NULL,
                           hThisInstance,
                           NULL);

    ShowWindow(hwnd, nCmdShow);

    while (GetMessage(&messages, NULL, 0, 0))
    {
        TranslateMessage(&messages);
        DispatchMessage(&messages);
    }

    return (int)messages.wParam;
}
